#ifndef __SYNC_CRITICAL_RESOURCE_H_INCLUDED__
#define __SYNC_CRITICAL_RESOURCE_H_INCLUDED__

#include "SyncCriticalSection.h"

namespace sync
{

// Critical Resource is a class which control access to resource with critical section.
// Any access to resource will cause calling of enter() and when access finished leave()
// of critical section associated with resource.
// TCriticalResource is critical section based alternate for Interlocked functions family.
template<typename TResourceType>
class TCriticalResource
{
private:
   // Resource lockers is used to enter() and leave() critical sections.
   // Temporary objects of TResourceLocker will be created on stack by calling of get() 
   // and will control access to resource
   class TResourceLocker;
   class TConstResourceLocker;

public:
   TCriticalResource();
   explicit TCriticalResource(const TResourceType&);
   TCriticalResource(const TCriticalResource<TResourceType>&);
   TCriticalResource<TResourceType>& operator=(const TCriticalResource<TResourceType>&);

   // Get method allow user to get controled access to resource. It is unsafe to use get() method of
   // more than one TCriticalResource in one expression because it can be potential dead lock.
   TResourceLocker operator->();
   TResourceLocker operator*();

   TConstResourceLocker operator->() const;
   TConstResourceLocker operator*() const;

   // It is dead-lock unsafety to use many critical resources in one expression, but it is safety
   // to use many copies of resources there if any critical resources used.
   TResourceType copy() const;

   sync::CCriticalSection& section();
private:
   TResourceType _resource;
   mutable sync::CCriticalSection _access;
}; // class TCriticalResource

// Allow controled access to resource.
template<typename TResourceType>
class TCriticalResource<TResourceType>::TResourceLocker
{
public:
   TResourceLocker(TResourceType&, sync::CCriticalSection&);
   TResourceLocker& operator=(const TResourceType&);

   // This operator is safe because thread will recursive enter in one critical section
   TResourceLocker(const TResourceLocker&);
   ~TResourceLocker() throw();

   // Resource can be safe used by using of this method, because it object of 
   // TResourceLocker already constructed so critical section already entered
   operator TResourceType&() const;
   TResourceType& get() const;

   // This operators is neccessary, so TCriticalResource::operator-> can return
   // TResourceLocker object by value
   TResourceLocker* operator->();
   TResourceLocker* operator*();

private:
   // Unsafe operator. It's potential dead lock to use in one expression more that one resource locker
   TResourceLocker& operator=(const TResourceLocker&);

   TResourceType* _resource;
   sync::CCriticalSection* _access;
}; // class TResourceLocker

// Allow controled access to const resource.
template<typename TResourceType>
class TCriticalResource<TResourceType>::TConstResourceLocker
{
public:
   TConstResourceLocker(const TResourceType&, sync::CCriticalSection&);
   TConstResourceLocker(const TConstResourceLocker&);

   ~TConstResourceLocker() throw();

   // Resource can be safe used by using of this method, because it object of 
   // TConstResourceLocker already constructed so critical section already entered
   operator const TResourceType&() const;
   const TResourceType& get() const;

   // This operators is neccessary, so TCriticalResource::operator-> can return
   // TConstResourceLocker object by value
   TConstResourceLocker* operator->();
   TConstResourceLocker* operator*();
private:
   // Unsafe operator. It's potential dead lock to use in one expression more that one resource locker
   TConstResourceLocker& operator=(const TConstResourceLocker&);

   const TResourceType* _resource;
   sync::CCriticalSection* _access;
}; // class TConstResourceLocker

//
// Reference specialization
//

template<typename TResourceType>
class TCriticalResource<TResourceType&>
{
private:
   class TResourceLocker;
   class TConstResourceLocker;
public:
   explicit TCriticalResource(TResourceType&);
   TResourceLocker operator->();
   TResourceLocker operator*();

   TConstResourceLocker operator->() const;
   TConstResourceLocker operator*() const;
   TResourceType copy() const;

   sync::CCriticalSection& section();
private:
   TCriticalResource(const TCriticalResource<TResourceType&>&);
   TCriticalResource<TResourceType&>& operator=(const TCriticalResource<TResourceType&>&);
   TResourceType* _resource;
   mutable sync::CCriticalSection _access;
}; // class TCriticalResource<TResourceType&>

template<typename TResourceType>
class TCriticalResource<TResourceType&>::TResourceLocker
{
public:
   TResourceLocker(TResourceType&, sync::CCriticalSection&);
   TResourceLocker& operator=(const TResourceType&);
   TResourceLocker(const TResourceLocker&);
   ~TResourceLocker() throw();
   operator TResourceType&() const;
   TResourceType& get() const;
   TResourceLocker* operator->();
   TResourceLocker* operator*();
private:
   TResourceLocker& operator=(const TResourceLocker&);
   TResourceType* _resource;
   sync::CCriticalSection* _access;
}; // class TResourceLocker<TResourceType&>

template<typename TResourceType>
class TCriticalResource<TResourceType&>::TConstResourceLocker
{
public:
   TConstResourceLocker(const TResourceType&, sync::CCriticalSection&);
   TConstResourceLocker(const TConstResourceLocker&);
   ~TConstResourceLocker() throw();
   operator const TResourceType&() const;
   const TResourceType& get() const;
   TConstResourceLocker* operator->();
   TConstResourceLocker* operator*();
private:
   TConstResourceLocker& operator=(const TConstResourceLocker&);
   const TResourceType* _resource;
   sync::CCriticalSection* _access;
}; // class TConstResourceLocker<TResourceType&>

} // namespace sync

#endif // __SYNC_CRITICAL_RESOURCE_H_INCLUDED__